# encoding: utf-8

"""
Generic ANN-generator.
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


from ann_activation_functions import *
from ann_unsupervised_learning import *
from ann_supervised_learning import *
from numpy import *
from random import *
from copy import deepcopy

#todo: prettyprinte lister uten komma

#learningmode i layers
#eye-full og inverse eye topologier (øker med (max - min)/2 + 0.1)

#Float printing syntax
DECIMALS='%3.4f'

#Output lots of junk when True, may be useful for debugging
VERBOSE=False

#Output nothing. Should probably be set to False, unless running on robot:
SILENT=False

def create_ann(path):
	"""Reads ann parameters from text file, and returns initialized ann-object"""
	#initialize parameter arrays:
	links=[]
	layers={}
	globalparams={}
	
	#get parameters from settings-file:
	try:
		f= open(path, 'r')	
		for line in f.read().split('\n'):
			if len(line)>0:
				if line.isspace() or line[0]=='#':
					continue
				else:
					line_t=line.split(':')
					#check if parameter is global or not:
					if (len(line_t)>1):
						line=line_t[1].split('=')
						#locals()[line[0]]=(eval(line[1]))
						globalparams[line[0]]=(eval(line[1]))
					else:
						params=line.split(' ')
						item={}
						for p in params:
							p=p.split('=')
							if len(p)>1:
								item[p[0]]=eval(p[1])
						if (params[0]=='layer'):
							layername=item['layername']
							layers[layername]=item
						elif (params[0]=='link'):
							links.append(item)
						else:
							print "Wrong format of ann-settings file"
							exit(1)
	except IOError as e:
		print "IO Error:", e
		sys.exit(1)
	return ANN(layers, links, **globalparams)


class ANN:
	def __init__(self, layers, links, learningrate, exec_order, trainingrounds, maxrounds):
		self.layers=[]
		self.links=[]
		self.trainingrounds=trainingrounds
		self.learningrate=learningrate
		self.trainingsets=[]
		self.testsets=[]
		for name in exec_order:
			self.layers.append(Layer(maxrounds, True, **layers[name]))

		for l in links:
			prelayer=self.get_layer(l['prelayer'])
			postlayer=self.get_layer(l['postlayer'])
			l['prelayer']=prelayer
			l['postlayer']=postlayer
			link=Link(self.learningrate, **l)
			prelayer.add_outputlink(link)
			postlayer.add_inputlink(link)
			self.links.append(link)
	def __str__(self):
		"""Returns string representation of self."""
		string="--ANN--\n"
		string+="Learningrate: " + str(self.learningrate) + "\n"
		string+="Rounds of training: " + str(self.trainingrounds) + "\n"
		string+="LAYERS:" + "\n"
		string+=str(self.layers) + "\n"
		string+="LINKS:" + "\n"
		string+=str(self.links)
		return string		
	def __repr__(self):
		return str(self)
	def copy(self):
		"""Returns deepcopy of self"""
		return deepcopy(self)
	def get_layer(self, layername):
		"""Return layer object with given name, used to set pre- and postlayer for links"""
		for i in range(len(self.layers)):
			if self.layers[i].name==layername:
				return self.layers[i]
	def print_solution(self):
		"""Prints activation levels in output layer"""
		print [DECIMALS % node.activation_level for node in self.layers[-1].nodes]
	def print_activation_levels(self):
		"""Prints activation levels in all layers, in order of execution"""
		for l in self.layers:
			print [DECIMALS % node.activation_level for node in l.nodes]
			print [DECIMALS % node.membrane_potential for node in l.nodes]
		print '\n'
	def reset_layers_full(self):
		"""run reset_full() on all layers"""
		for i in range(len(self.layers)):
			self.layers[i].reset_full()
	def get_output(self):
		return [node.activation_level for node in self.layers[-1].nodes]	
	def train(self, trainingsets):
		self.trainingsets=trainingsets

		print "####TRAINING####"
		
		for trainingset in self.trainingsets:
			for i in range(self.trainingrounds):
				self.reset_layers_full()
				if VERBOSE: print "Training set" + str(i)
				self.layers[0].initialize_inputs(trainingset)
				for l in range(len(self.layers)):
					if VERBOSE: print "executing layer" + str(l) + "..."
					self.layers[l].run()
				for l in range(len(self.links)):
					if VERBOSE: print "run learning on link from " + str(self.links[l].prelayer.name) + " to " + str(self.links[l].postlayer.name) + "..."
					self.links[l].learn()
				#print self.links[0]
				#print self.links[10]
			self.print_activation_levels()
	def test(self, testsets):
		self.testsets=testsets
		
		if not SILENT: print "####TESTING####"
		for testset in self.testsets:
			self.reset_layers_full()
			self.layers[0].initialize_inputs(testset)
			for l in range(len(self.layers)):
				if VERBOSE: print "executing layer" + str(l) + "..."
				self.layers[l].run()
				#print self.links[3]
			if not SILENT: self.print_activation_levels()
			
		

class Layer():
	def __init__(self, maxrounds, learning, layername, neurons, activationFunction, tactive, tquiescent, ractive, rquiescent, threshold=0.0):
		self.name=layername
		self.n_nodes=neurons
		self.activationFunction=activationFunction(threshold)
		self.max_settling_rounds=maxrounds
		self.active_testing=ractive
		self.active_training=tactive
		self.quiescent_training=tquiescent
		self.quiescent_testing=rquiescent
		
		#If set, layer can update neuron activation levels and send signals to downstream neurons:
		self.active=self.active_training
		#If set, layer runs to quiescence:
		self.quiescent_mode=self.quiescent_training
		#If set, input links is a
		self.learning_mode=learning
		
		self.changed=0
		
		self.inputlinks=[]
		self.outputlinks=[]
		self.nodes=[]
		
		for i in range(self.n_nodes):
			self.nodes.append(Node(i, self))
	def __str__(self):
		"""Returns string representation of self."""
		string="Layer: " + str(self.name) + " - "
		string+="Neurons: " + str(self.nodes) + "\n"
		return string
	def __repr__(self):
		return str(self)
	def copy(self):
		"""returns deepcopy of self"""
		return deepcopy(self)
	def initialize_inputs(self, data):
		for i in range(len(data)):
			self.nodes[i].set_membrane_potential(float(data[i]))
			self.nodes[i].set_activation_level(float(data[i]))
			self.nodes[i].set_prev_activation_level()
	def reset(self):
		self.changed=0
		for i in range(len(self.nodes)):
			self.nodes[i].reset()
	def reset_full(self):
		self.changed=0
		for i in range(len(self.nodes)):
			self.nodes[i].reset_full()
	def set_test_mode(self):
		self.active=self.active_testing
		self.quiescent=self.quiescent_testing
	def print_node_values(self):
			print "act.levels", [DECIMALS % node.activation_level for node in self.nodes]
			print "memb. pots", [DECIMALS % node.membrane_potential for node in self.nodes]
			print "prev act.levels", [DECIMALS % node.prev_activation_level for node in self.nodes]
	def get_activation_levels(self):
			return [DECIMALS % node.activation_level for node in self.nodes]
	def run(self):
		if self.active:
			if self.quiescent_mode:
				rounds=self.max_settling_rounds
			else:
				rounds=1
			
			if self.name!=0:
				self.reset()
			for i in range(rounds):
				self.changed=0
				self.update_node_values()
				self.set_prev_activation_levels()
				if VERBOSE: self.print_node_values()
				if self.changed==0:
					if VERBOSE: print "Layer " + str(self.name) + ", iteration " + str(i) + ": No change, breaking.."
					break
	
	def add_inputlink(self, link):
		self.inputlinks.append(link)
	def add_outputlink(self, link):
		self.outputlinks.append(link)
	def update_node_values(self):
		"""Calculate membrane potentials and activation levels for all nodes in layer"""
		for i in range(len(self.nodes)):
			sum=0.0
			if self.name==0:
				sum=self.nodes[i].membrane_potential
			else:
				for link in self.inputlinks:
					if link.prelayer.active==True:
						arcs=link.getinputarcs(i)
						for a in arcs:
							sum+=a.weight*a.prenode.prev_activation_level
				if VERBOSE: print "Total node sum: ", sum
			if (sum!=self.nodes[i].membrane_potential):
				self.changed+=1	
			self.nodes[i].set_membrane_potential(sum)
			self.nodes[i].set_activation_level(self.activationFunction.activate(sum))
	def set_prev_activation_levels(self):
		for i in range(len(self.nodes)):
			self.nodes[i].set_prev_activation_level()


class Node():
	def __init__(self, number, layer):
		self.layer=layer
		self.number=number
		#Sum of weighted inputs:
		self.membrane_potential=0.0
		#Node output:
		self.activation_level=0.0
		self.prev_activation_level=self.activation_level
	def __str__(self):
		"""Returns string representation of self."""
		string=str(DECIMALS % self.activation_level)
		return string
	def __repr__(self):
		return str(self)
	def copy(self):
		"""returns deepcopy of self"""
		return deepcopy(self)
	def set_membrane_potential(self,new_membrane_potential):
		self.membrane_potential=new_membrane_potential
	def set_activation_level(self,new_activation_level):
		self.activation_level=new_activation_level
	def set_prev_activation_level(self):
		self.prev_activation_level=self.activation_level
	def reset(self):
		self.membrane_potential=0.0
		self.activation_level=0.0
	def reset_full(self):
		self.reset()
		self.prev_reset()	
	def prev_reset(self):
		self.prev_activation_level=0.0		

class Link():
	def __init__(self, learningrate, prelayer, postlayer, topology, maxweight, minweight, learningrule=None, plastic=None, weights_from_file=None):
		self.prelayer=prelayer
		self.postlayer=postlayer
		self.topology=topology
		self.minweight=minweight
		self.maxweight=maxweight
		self.learningrate=learningrate
		if learningrule!=None: self.learningfunction=learningrule(learningrate)
		self.plastic=plastic
		
		self.initialize_arcs()
		if weights_from_file: self.set_weights_from_file(weights_from_file)
		
	def __str__(self):
		"""Returns string representation of self."""
		string="Link from layer" + str(self.prelayer.name) + " to " + str(self.postlayer.name) + "\n"
		string+="Arcs:\n" + str(self.arcs) + "\n"
		return string
	def __repr__(self):
		return str(self)
	def copy(self):
		"""Returns deepcopy of self"""
		return deepcopy(self)
	def initialize_arcs(self):
		self.arcs=empty((self.postlayer.n_nodes, self.prelayer.n_nodes), dtype=object)
		if self.topology=='full':
			for i in range(self.prelayer.n_nodes):
				for j in range(self.postlayer.n_nodes):
					self.arcs[j,i]=Arc(self.prelayer.nodes[i],self.postlayer.nodes[j], uniform(self.minweight,self.maxweight), self)
		elif self.topology=='1-1':
			for i in range(self.prelayer.n_nodes):
				for j in range(self.postlayer.n_nodes):
					if (i==j):
						self.arcs[j,i]=Arc(self.prelayer.nodes[i],self.postlayer.nodes[j], uniform(self.minweight,self.maxweight), self)
		elif self.topology=='triangular':
			for i in range(self.prelayer.n_nodes):
				for j in range(self.postlayer.n_nodes):
					if (i!=j):
						self.arcs[j,i]=Arc(self.prelayer.nodes[i],self.postlayer.nodes[j], uniform(self.minweight,self.maxweight), self)
		elif self.topology=='eye':
			for i in range(self.prelayer.n_nodes):
				for j in range(self.postlayer.n_nodes):
					print "todo: not yet implemented"
					exit(1)
					
	def getinputarcs(self, number):
		arcs=[]
		for i in range(self.prelayer.n_nodes):
			if self.arcs[number, i]!=None:
				arcs.append(self.arcs[number, i])
		return arcs
	def learn(self):
		if (self.plastic):
			for i in range(len(self.prelayer.nodes)):
				for j in range(len(self.postlayer.nodes)):
					arc=self.arcs[j,i]
					if arc!=None:
						self.arcs[j,i].update(self.learningfunction.update(arc.prenode.activation_level, arc.postnode.activation_level, arc.weight))
	def modify_weights(self, dwarray):
		for i in range(len(self.prelayer.nodes)):
			for j in range(len(self.postlayer.nodes)):
				if self.arcs[j,i]!=None:
					self.arcs[j,i].update(dwarray[j,i])
	def set_weights_from_file(self, filename):
		"""Load 2d array from 'filename' into link weights.
		"""
		warray=loadtxt(filename)
		for i in range(len(self.prelayer.nodes)):
			for j in range(len(self.postlayer.nodes)):
				if self.arcs[j,i]!=None:
					self.arcs[j,i].weight=warray[j,i]
	def save_weights_to_file(self, filename):
		"""Save weight configuration to file.
		"""
		warray=empty((self.postlayer.n_nodes, self.prelayer.n_nodes), dtype=float)
		for i in range(len(self.prelayer.nodes)):
			for j in range(len(self.postlayer.nodes)):
				if self.arcs[j,i]!=None:
					warray[j,i]=self.arcs[j,i].weight
				else:
					warray[j,i]=0.0
		savetxt(filename,warray)
	
class Arc:
	def __init__(self, prenode, postnode, initial_weight, link):
		self.prenode=prenode
		self.postnode=postnode
		self.weight=initial_weight
		self.initial_weight=initial_weight
		self.link=link
	def __str__(self):
		"""Returns string representation of self."""
		string= str(DECIMALS % self.weight)
		return string 
	def __repr__(self):
		return str(self)
	def copy(self):
		"""Returns deepcopy of self"""
		return deepcopy(self)
	def update(self, dw):
		"""Updates the current weight with dw"""
		self.weight+=dw
		

