# encoding: utf-8

"""
Methods for setting up a behavior based robot controller.
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from __future__ import with_statement
import threading
import time
import signal

from utilities import *
from image import *


#To avoid problems with epuck.talk responses and other annoying side-effects
#of threading:
#lock=threading.Lock()

#Prints debugging information:
VERBOSE=True

#If silent, no output is given. Should probably be left false:
SILENT=False

#height of behavior tree, 5 should be enough for now
BEHAVIOR_PRIORITY_LEVELS=5

#size of historybuffer
HISTORYLENGTH=10

#timestep for epuck movements
TIMESTEP=0.25

#decorator for using RLock on function:
def synchronized(self):
    lock = threading.RLock()
    def fwrap(function):
        def new_function(*args, **kw):
            with lock:
                return function(*args, **kw)
        return new_function
    return fwrap


class BehaviorController(threading.Thread):
	"""Main robot controller class. All other controllers should inherit
	from this class, but should implement their own version of run().
	"""
	def __init__(self, epuck, behaviors, lock, threaded):
		if threaded: threading.Thread.__init__(self)
		#current orientation relative to starting orientation: 
		self.orientation=0.0
		self.collision_counter=0
		self.behaviors=behaviors
		self.motor_speeds=[]
		self.epuck=epuck
		self.lock=lock
		self.threaded=threaded
		self.sleeptime=0.1
		self.historybuffer=CircularBuffer(HISTORYLENGTH)
		for i in range(BEHAVIOR_PRIORITY_LEVELS):
			self.motor_speeds.append([])
	def initialize(self):
		"""Initializes controller and behaviors. Supports parallell and 
		sequential mode.
		"""
		if self.threaded:
			for b in self.behaviors:
				b.controller=self
				b.start()
			self.start()
			#for b in self.behaviors:
				#b.join()
		else:
			#scaling factor, translates sleeptime into iteration steps. system dependent
			f=20
			for b in self.behaviors:
				b.controller=self
				#adjust sleeptime
				b.sleeptime=f*b.sleeptime
			for i in range(100000):
				for b in self.behaviors:
					b.controller=self
					#use adjusted sleeptime to determine when to run:
					if i%(b.sleeptime)==0:
						b.run()
				self.run()
			
	def run(self):
		"""Command arbitration/fusion
		"""
		while True:
			self.lock.acquire()
			if VERBOSE: print "Motor speeds:", self.motor_speeds
			try:
				for ms in self.motor_speeds:
					if nonzero(ms):
						if VERBOSE: print "Motor speeds:", ms
						if len(ms)>0:
							dleft, dright=list_element_sum(ms)
							if VERBOSE: print "Changing motor speeds: ", dleft, dright
							self.epuck.set_wheel_speeds(dleft,dright, TIMESTEP)
							self.orient(left_speed=dleft, right_speed=dright)
							print "orientation: ", self.orientation
							break
				for i in range(BEHAVIOR_PRIORITY_LEVELS):
					self.motor_speeds[i]=[]
				self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except OSError as e:
				print "Should probably not happen:", e
				sys.exit()
	def kill(self):
		"""Kill thread, and notify mother-process.
		"""
		print "Thread killed"
		try:
			self.epuck.death()
			os.kill(self.child, signal.SIGKILL)
		except OSError: pass				
	def count_collisions(self):
		"""Count the number of times one of the IR sensors is >3000
		"""
		if max(self.epuck.proximity_sensors)>3000:
			self.collision_counter+=1
	def pause_all(self, s):
		"""behavior s pauses all others except itself
		"""
		for b in self.behaviors:
			if b!=s:
				b.active=False
	def activate_all(self):
		"""Unpause all running behaviors
		"""
		for b in self.behaviors:
			b.active=True
	def pause(self, behaviorid):
		"""Pause given behavior
		""" 
		for i in range(len(self.behaviors)):
			if self.behaviors[i].id==behaviorid:
				self.behaviors[i].active=False
				break
	def activate(self, behaviorid):
		"""Activate given behavior
		"""
		for i in range(len(self.behaviors)):
			if self.behaviors[i].id==behaviorid:
				self.behaviors[i].active=True	
				break
	def orient(self, new_orientation=None, left_speed=None, right_speed=None):
		"""Update current orientation, based on wheel movement and/or waypoints
		 1 0 7   UP
		 2	X 6   ^
	    3 4 5   | 
	    
	    angle a, orientation o:
	    a/45=o 
	    7 rounds with speed 0.5 = 6 -> coefficient 6/7
		"""
		if new_orientation!=None:
			self.orientation=new_orientation
		elif left_speed!=None and right_speed!=None:
			if left_speed==-right_speed:
				print "spin!:", left_speed, right_speed
				self.orientation=(self.orientation + right_speed/0.5*(6.0/7.0)) % 8.0
				
			
			
class Behavior(threading.Thread):
	"""Behavior superclass
	"""
	def __init__(self, epuck, lock,threaded,priority):
		if threaded: threading.Thread.__init__(self)
		self.epuck=epuck
		self.controller=None
		self.priority=priority
		self.lock=lock
		self.sleeptime=1
		self.historybuffer=CircularBuffer(HISTORYLENGTH)
		
		#All behaviors start out active, with the possibility of being inhibited
		#by others
		self.active=True
		
		#If true, behavior assumes it is run as a separate thread.
		self.threaded=threaded
		
		#see comments in watch and kill-methods. 
		#Creates a child thread, which returns. The parent thread waits for a 
		#KeyboardInterrupt and then kills off the child thread:
		if self.threaded:
			self.child = os.fork()
			if self.child == 0:
				return
			else:
				self.watch()		
	def run(self):
		"""run() is implemented in child classes
		"""
		return NotImplemented
	def watch(self):
		"""watch and kill makes sure keyboard interrupts are handled by the thread. 
		Code borrowed from The Little Book of Semaphores [http://greenteapress.com/semaphores/]
		"""
		try:
			os.wait()
		except KeyboardInterrupt:
			print 'KeyboardInterrupt'
			self.kill()
		sys.exit()
	def kill(self):
		"""watch and kill makes sure keyboard interrupts are handled by the thread. 
		Code borrowed from The Little Book of Semaphores [http://greenteapress.com/semaphores/]
		"""
		print "Thread killed"
		try:
			self.epuck.death()
			os.kill(self.child, signal.SIGKILL)
		except OSError: pass
		
class Wander(Behavior):
	def __init__ (self,epuck, ann, leftspeed, rightspeed, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=BEHAVIOR_PRIORITY_LEVELS-1)
		self.ann=ann
		self.left = leftspeed
		self.right= rightspeed
		self.sleeptime=5
		self.id=0
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "wander"
					dv=random_float(-0.5,0.5)
					#dright=random_float(-0.5,0.5)
					if VERBOSE: print "Motor speeds dleft, dright", dv, -dv
					#self.epuck.change_motor_speeds(dleft,dright, TIMESTEP)
					self.controller.motor_speeds[self.priority].append([dv,-dv])
					print self.controller.motor_speeds
					if VERBOSE: print "end wander"
					self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in Wander: ", e
				self.kill()

class AvoidObstacles(Behavior):
	def __init__ (self,epuck, ann, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=0)
		self.ann=ann
		self.status = 1
		self.sleeptime=0.1
		self.id=1
		
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Obstacle avoidance"
					prox=self.epuck.get_proximities()
					#skip rounds with empty sensor output (happens when battery level is low)
					if len(prox)>1:	
						if not SILENT: print "IR sensors: ", prox	
						self.ann.test([prox])
						dleft, dright =self.ann.get_output()
						self.controller.motor_speeds[self.priority].append([dleft/2.0,dright/2.0])
						if VERBOSE: print "End obstacle avoidance"
					self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in AvoidObstacles: ", e
				self.kill()

class GotoSound(Behavior):
	def __init__ (self,epuck, ann, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=0.5
		self.id=2
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					print self.lock.locked()
					if VERBOSE: print "Go to source of sound"
					mics=self.epuck.get_mic_amplitudes()
					#skip rounds with empty sensor output (happens when battery level is low):
					if len(mics)>1:	
						if not SILENT: print "Mic amplitudes: ", mics	
						self.ann.test([mics])
						dleft, dright =self.ann.get_output()
						self.controller.motor_speeds[self.priority].append([dleft/2.0,dright/2.0])
						if VERBOSE: print "End goto sound"
					self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GotoSound: ", e
				self.kill()
				
class GotoLight(Behavior):
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=2
		self.id=3
		self.light_threshold=10
		self.found_light_threshold=250
	def run(self):
		while True:
			try:
				success=False
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Begin search for light"
					#Pause GoHome:
					self.controller.pause(4)
					
					#stop robot
					self.epuck.move_wheels(0.0,0.0)
					#rotate approx. 306 degrees in 4 steps:
					for i in range(4):
						print "i:",i
						im=self.epuck.snapshot()
						#returns list with length 3:
						white_pixels=pixel_count_band(im,0,30,13)
						if not SILENT: print "Goto light pixels: ", white_pixels
						
						if sum(white_pixels)>self.found_light_threshold: 
							if not SILENT: print "Found light"
							if not SILENT: im.show()
							success=True
							break
						elif max(white_pixels) > self.light_threshold:
							dleft,dright=0.5,0.5
						else:
							dleft,dright=0.0,0.0
						if (dleft!=0.0 or dright!=0.0): 
							self.controller.motor_speeds[self.priority].append([dleft,dright])
							break
						else:
							self.epuck.move_wheels(-1.0,1.0,TIMESTEP) 
							
					#pause self, activate GoHome:
					if success:
						self.controller.pause(3)
						self.controller.activate(4)
					self.epuck.move_wheels(0.5,0.5)
					
					if VERBOSE: print "End search for light"
					self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GotoLight: ", e
				self.kill()
		
class GoHome(Behavior):
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		#this behavior should start out inactive
		self.active=False
		self.status = 1
		self.sleeptime=2
		self.id=4
		self.redness_threshold=50
		self.found_home_threshold=1000
	def run(self):
		while True:
			try:
				success=False
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Begin Go home"
					#Pause GotoLight:
					self.controller.pause(3)
					
					#stop robot
					self.epuck.move_wheels(0.0,0.0)
					
					#rotate approx. 306 degrees in 4 steps:
					for i in range(4):
						if not SILENT: print "i:",i
						im=self.epuck.snapshot()
						#returns list with length 3:
						red_pixels=redness_count_band(im,0,30,13)
						if not SILENT: print "Go home pixels: ", red_pixels
						if sum(red_pixels)>self.found_home_threshold: 
							if not SILENT: print "found home"
							if not SILENT: im.show()
							success=True
							break
						elif max(red_pixels)> self.redness_threshold:
							dleft,dright=0.5,0.5
							
							break
						else:
							dleft,dright=0.0,0.0
								
						if (dleft!=0.0 or dright!=0.0): 
							self.controller.motor_speeds[self.priority].append([dleft,dright])
							break
						else:
							self.epuck.move_wheels(-1.0,1.0,TIMESTEP) 
				
					#pause self, activate GoHome:
					if success:
						self.controller.pause(3)
						self.controller.activate(4)
					self.epuck.move_wheels(0.5,0.5)
					
					if VERBOSE: print "End Go home"
					self.lock.release()
				
				if not self.threaded: break
				time.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GoHome: ", e
				self.kill()
				
class FollowLine(Behavior):
	"""incomplete..
	"""
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=2
		self.id=5
	def run(self):
		return NotImplemented
