#!/usr/bin/env python
# encoding: utf-8

"""
E-puck communication methods (requires use of pucker).
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


#E-puck controller:
import pucker
import sys
import serial

#Used for interactive():
import curses

#Print debugging information:
VERBOSE=False

class Epuck(pucker.epuck):
	"""Overrides some of the methods in the pucker framework, and inherits the rest"""
	def __init__(self, port="/dev/rfcomm0", id=0, *args, **kwargs):
		pucker.epuck.__init__(self, id, port, *args, **kwargs)

	def move_wheels(self, left = 0.0, right = 0.0, duration = None):
		"""General move routine for forward or backward motion. Speeds are numbers
		between -1 and 1, these are converted to values in the range [-1000,1000] 
		before calling set_motor_speeds. Can either be time-limited (by setting duration),
		or permanent.
		"""
		sv = self.maxvel * self.robot_maxvel
		ls = max(-1.0, min(1.0, left)) * sv
		rs = max(-1.0, min(1.0, right)) * sv
		self.set_motor_speeds(left = ls, right = rs)
		if duration:
			self.wait(duration)
			self.stop_moving()
	def change_motor_speeds(self, dleft=0.0, dright=0.0, duration=None):
		"""Changes the motorspeeds by factors dleft and dright (should be between 
		-1 and 1). Can either be time limited (by setting duration), or permanent).
		"""		
		speeds =self.get_motor_speeds()
		#print "speeds", speeds
		while (len(speeds)!=2):
			print "Incomplete motor speeds. Epuck battery probably needs recharging. Will keep trying to fetch speed data", speeds
			speeds =self.get_motor_speeds()
		
		print "Changing motor speeds, was:", speeds
		ls, rs=speeds
		
		scale = self.maxvel * self.robot_maxvel
		left= max(-1.0, min(1.0, ls/scale))
		right= max(-1.0, min(1.0, rs/scale))
		if VERBOSE: print "New speed (left,right):", left+dleft,right+dright, ", was", left, right
		self.move_wheels(dleft,dright,duration)
		self.move_wheels(left, right)
	def death(self):
		"""Stop and disconnect epuck
		"""
		try:
			if self.simob:
				self.simob.remove_robot(self.id)
			else:
				self.stop_moving()
				self.disconnect()
		except serial.serialutil.SerialException as e:
			#e-puck is already disconnected
			pass
			
	#@com_wrap
	def snapshot(self,xres = False, yres = False, show=False):
		"""Return 40x40 image if nothing else is specified. Optional 
		whether show() is called or not. 
		"""
		try:
			if self.simob:
				im = self.simob.capture_image(self.id)
			else:
				im = self.robot_snapshot(xres = xres, yres = yres)
			if im and show:
				im.show()
			if im:
				return im
		except:
			print "error in snapshot()"
			
	def interactive(self):
		"""Another version of run_toy(). The wheels are controlled by arrow keys.
		Will probably break on windows...
		"""
		try:
			stdscr = curses.initscr()
			curses.cbreak()
			stdscr.keypad(1)
			curses.echo()
			#img-counter
			i=0
			while True:
				c = stdscr.getch()
				if c==ord('s'):
					im=self.snapshot()
					imgname="snapshot" + str(i) + ".jpg"
					im.save(imgname)
					i+=1
				elif c==ord('q'):
					raise KeyboardInterrupt()
				elif c==ord('p'):
					print self.get_proximities()
				elif c==ord('m'):
					print self.get_mic_amplitudes()
				elif c==curses.KEY_RIGHT:
					self.move_wheels(1.0,-1.0,0.25)
				elif c==curses.KEY_LEFT:
					self.move_wheels(-0.5,0.5,0.25)
				elif c==curses.KEY_UP:
					self.move_wheels(0.5,0.5,0.25)
				elif c==curses.KEY_DOWN:
					self.move_wheels(-0.5,-0.5,0.25)	
		except KeyboardInterrupt as e:
			print "Quiting..: ", e
		except Error as e:
			print "unknown error.."
		finally:
			print "cleaning up.."
			self.death()
			curses.nocbreak(); stdscr.keypad(0); curses.echo()
			curses.endwin()
			sys.exit(0)

	
	##low level methods for communication with physical robot only:##	
	def get_mic_amplitudes(self):
		"""Return the amplitudes of microphones as a list: [right,left, back]
		"""
		if self.simob:
			print "get_mic_amplitudes() are not yet supported for simulated robots"
			sys.exit(1)
		else:
			return self.query_numbers("U")

