from controller import *
import math
import time
import sys,os
import Image


class Epuck_Webots (DifferentialWheels):
	def __init__(self, tempo = 0.5):
		DifferentialWheels.__init__(self)
		
		##PHYSICAL EPUCK PROPERTIES:##
		self.max_wheel_speed = 1000
		#Number of proximity sensors:
		self.n_prox_sensors = 8
		#for wheel encoders:
		self.encoder_resolution = 159.23
		#Camera resolution:
		self.camera_xres = 40
		self.camera_yres = 40
		#from EPuckSound.proto:
		self.wheel_diameter = 0.0205*2 #m
		self.axle_length = 0.052 #m
		
		#Real-time seconds per timestep:
		self.timestep_duration = 1 
		
		#Upper velocity bound = Fraction of the robot's maximum velocity = 1000 = 1 wheel revolution/sec
		self.tempo = tempo
		self.wheel_speed_scaling=self.max_wheel_speed*self.tempo
		#self.max_spin_rate = self.tempo * (self.wheel_diameter / self.axle_length) 
		self.left_speed=0.0
		self.right_speed=0.0
	
		#Fetch from worldinfo.basicTimeStep:
		self.timestep = int(self.getBasicTimeStep())
		
		#Setup sensors:
		self.enableEncoders(self.timestep)
		self.camera = self.getCamera('camera')
		self.camera.enable(4*self.timestep)
		self.prox_sensor_values = [0.0]*self.n_prox_sensors
		self.prox_sensors = [self.getDistanceSensor('ps'+str(x)) for x in range(self.n_prox_sensors)]
		map((lambda s: s.enable(self.timestep)), self.prox_sensors) # Enable all distance sensors
		
		#So that the keyboard can be used to control epuck:
		self.keyboardEnable(6*self.timestep)
		
		#Add emitter and receiver to epuck.
		self.emitter=self.getEmitter("emitter")
		self.receiver=self.getReceiver("receiver")
		self.receiver.enable(4*self.timestep)
		
	def wait(self, seconds = 1.0):
		"""Makes epuck sleeps for a given number of seconds
		"""
		print "waiting..."
		time.sleep(seconds)

	def set_wheel_speeds(self,left = 0.0, right = 0.0):
		"""Sets wheel speeds. left and right are floats in the range [-1,1],
		and is scaled by self.wheel_speed_scaling
		"""
		s = self.wheel_speed_scaling
		left=max(-1.0, min(1.0, left))
		right=max(-1.0, min(1.0, right))
		
		self.left_speed=left
		self.right_speed=right
		self.setSpeed(int(left*s),int(right*s))
	
	def change_wheel_speeds(self,dl=0.0,dr=0.0):
		"""Changes wheel speeds by factors dl and dr. 
		"""
		left=self.left_speed+dl
		right=self.right_speed+dr
		self.set_wheel_speeds(left,right)
		
# The angle is in degrees.  It must be converted to "encoder units" of the wheels, of which there are
# res per radians of wheel rotation (res = the differentialwheels.encoderResolution with values such as 100)
	def spin_angle(self, angle=0.0):
		old_left, old_right=self.left_speed, self.right_speed
		rad = abs(angle)*math.pi/180.0
		if angle < 0:
			self.set_wheel_speeds(-1.0,1.0)
		else: 
			self.set_wheel_speeds(1.0,-1.0)
		self.setEncoders(0.0,0.0)
		total_encoder_steps = rad*self.encoder_resolution*self.axle_length/self.wheel_diameter
		current_step = abs(self.getLeftEncoder())
		stop_step = current_step + total_encoder_steps
		while current_step < stop_step:
			self.step(self.timestep)
			current_step = abs(self.getLeftEncoder())
		self.set_wheel_speeds(old_left,old_right)

	def get_proximities(self):
		"""Store and return proximity sensor values, converted to integers.
		"""
		for i in range(self.n_prox_sensors):
			self.prox_sensor_values[i] = int(self.prox_sensors[i].getValue())
		return self.prox_sensor_values

	def get_image(self):
		"""Aquire string-image from epuck camera, and convert it to python image (using PIL)
		"""
		strImage=self.camera.getImage()
		im = Image.fromstring('RGB',(self.camera.getWidth(), self.camera.getHeight()), strImage)
		return im

	def interactive(self):
		"""Control epuck interactively, using the keyboard. 
		"""
		try:
			#img-counter
			i=0
			while True:
				c = self.keyboardGetKey()
				if c==ord('S'):
					im=self.get_image()
					imgname="snapshot" + str(i) + ".jpg"
					im.save(imgname)
					i+=1
				elif c==ord('Q'):
					break
				elif c==ord('P'):
					print self.get_proximities()
				elif c==ord('E'):
					print "emitting.."
					self.emitter.send("test")
				elif c==ord('R'):
					print "todo: getting receiver data"
					if self.receiver.getQueueLength()!=0:
						data=self.receiver.getData()
						print data
					else:
						print "nothing to receive"
					#print self.get_mic_amplitudes()
				elif c==Robot.KEYBOARD_RIGHT:
					self.spin_angle(25)
				elif c==Robot.KEYBOARD_LEFT:
					self.spin_angle(-25)
				elif c==Robot.KEYBOARD_UP:
					self.change_wheel_speeds(0.25,0.25)
				elif c==Robot.KEYBOARD_DOWN:
					self.change_wheel_speeds(-0.25,-0.25)
				
				if self.step(self.timestep) == -1: break
		except KeyboardInterrupt as e:
			print "Quitting..: ", e
		finally:
			pass

	def run_behaviors(self):
		pass


# This is the main program of your controller.
# It creates an instance of your Robot subclass, launches its
# function(s) and destroys it at the end of the execution.
# Note that only one instance of Robot should be created in
# a controller program.
epuck_controller = Epuck_Webots()
epuck_controller.interactive()
