#!/usr/bin/env python
# encoding: utf-8

"""
Activation functions, to be used to calculate activation levels for nodes in ANN's
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from numpy import *

class ActivationFunction:
	"""superclass of different activation functions"""
	def __init__(self, threshold=0.5):
		self.threshold=threshold
	def activate(self, membrane_potential):
		return NotImplemented

class Step(ActivationFunction):
	def __init__(self, threshold=0.5):
		ActivationFunction.__init__(self, threshold)
	def activate(self, membrane_potential):
		"""Output is 0 if potential is below threshold, and 1 if above
		"""
		if membrane_potential==self.threshold:
			output=0.5
		elif membrane_potential>self.threshold:
			output=1.0
		elif membrane_potential<self.threshold:
			output=0.0
		return output
	def derivative_activate(self, membrane_potential):
		return self.threshold

class VariableStep(ActivationFunction):
	"""Step function with variable high, low and threshold settings. 
	Usage: VariableStep threshold=[thresh,low,high]
	"""
	def __init__(self, param):
		threshold=param[0]
		self.low=param[1]
		self.high=param[2]
		ActivationFunction.__init__(self, threshold)
	def activate(self, membrane_potential):
		"""Output is low if potential is below threshold, and high if above
		"""
		if membrane_potential==self.threshold:
			output=(self.low+self.high)/2
		elif membrane_potential>self.threshold:
			output=self.high
		elif membrane_potential<self.threshold:
			output=self.low
		return output
	def derivative_activate(self, membrane_potential):
		return self.threshold

class Sign(ActivationFunction):
	def __init__(self, threshold=0.5):
		ActivationFunction.__init__(self, threshold)
	def activate(self, membrane_potential):
		"""Output is -1 if potential is below threshold,
		and 1 if above.
		"""
		if membrane_potential==self.threshold:
			output=0.0
		elif membrane_potential>self.threshold:
			output=1.0
		elif membrane_potential<self.threshold:
			output=-1.0
		return output

class Linear(ActivationFunction):
	def __init(self):
		ActivationFunction.__init__(self, None)
	def activate(self, membrane_potential):
		"""Output is the membrane potential
		"""
		return membrane_potential
		
class PositiveLinear(ActivationFunction):
	def __init(self):
		ActivationFunction.__init__(self, None)
	def activate(self, membrane_potential):
		"""Output is the membrane potential, 
		or 0 if membrane potential is negative. 
		"""
		v=0.0
		if membrane_potential>0.0:
			v=membrane_potential
		return v
	
class Tanh(ActivationFunction):
	def __init(self):
		ActivationFunction.__init__(self, None)
	def activate(self, membrane_potential):	
		"""Returns tanh.
		"""
		return tanh(membrane_potential)
	def derivative_activate(self, membrane_potential):
		return 1.0 - tanh(membrane_potential)**2
		
class Sigmoid(ActivationFunction):
	def __init(self):
		ActivationFunction.__init__(self, None)
	def activate(self, membrane_potential):	
		"""Returns sigmoid, normalized to [0, 1].
		"""
		return 1.0/(1.0+exp(-membrane_potential))	
	def derivative_activate(self, membrane_potential):
		return exp(-membrane_potential)/(1+exp(-membrane_potential))**2
		
class HyperbolicSigmoid(ActivationFunction):
	"""Should accelerate learning, a,b -> guyon 1991)
	"""
	def __init(self):
		ActivationFunction.__init__(self, None)
		self.a=1.716
		self.b=0.667
	def activate(self, membrane_potential):	
		"""
		Returns sigmoid, normalized to [0, 1].
		"""
		return (2*self.a)/(1.0+exp(-self.b*membrane_potential)) - self.a
