# encoding: utf-8

"""
Learning algorithms to be used for supervised neural network learning.
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

VERBOSE=False
SILENT=False

def backpropagation(ann, trainingpairs, maxerror):
	"""Backpropagation learning algorithm, used on feed-forward neural networks.
	ann should be an ANN-object, trainingpairs a tuple of input and desired patterns.
	ANN is assumed to be pure feed-foreward.
	"""
	
	eta=ann.learningrate
	
	if not SILENT: print "####Running backpropagation algorithm...####"

	rounds=0
	error=1
	while rounds<ann.trainingrounds:
		rounds+=1
		if not SILENT: 
			if (rounds % 100 == 0): print "Round ", rounds, "error", error
		
		#break when error is sufficiently small
		if error<maxerror: 
			print "Breaking after round ", rounds, "error=", error
			break
		error=0.0
		
		for trainingpair in trainingpairs:
			inputpattern, desiredpattern = trainingpair
			if VERBOSE: print "Training set (input, desired output):", inputpattern, desiredpattern
			ann.layers[0].initialize_inputs(inputpattern)
		
			#run through network:
			for i in range(len(ann.layers)):
					if VERBOSE: print "executing layer" + str(i) + "..."
					ann.layers[i].run()
					
			##Calculate deltas:##
			#backwards iteration:
			for l in range(-1,-len(ann.layers)-1, -1):
				current_layer=ann.layers[l]
				
				if len(current_layer.inputlinks)>1 or len(current_layer.outputlinks)>1:
					print "Layer has more than one input or outputlink. Aborting.."
					sys.exit(1)
				
				for j in range(current_layer.n_nodes):
					#calculate derivative of the activation function of node j:
					dactivation=current_layer.activationFunction.derivative_activate(current_layer.nodes[j].membrane_potential)
					
					#if current layer is output-layer:
					if l==-1:
						delta=dactivation*(desiredpattern[j] - current_layer.nodes[j].activation_level)
					else:
						#sum over k's(prevdelta*weight)
						#assumes only one outputlink in current layer:
						current_outputlink=current_layer.outputlinks[0]
						next_layer=current_outputlink.postlayer
						deltasum=0.0
						for k in range(next_layer.n_nodes):
							deltasum+=next_layer.nodes[k].delta*current_outputlink.arcs[k,j].weight
						delta=dactivation*deltasum
						
					current_layer.nodes[j].delta=delta
			
			##Update weights:##
			#backwards iteration:
			for l in range(-1,-len(ann.layers), -1):
				current_layer=ann.layers[l]
				current_inputlink=current_layer.inputlinks[0]
				prev_layer=current_inputlink.prelayer
				
				for i in range(prev_layer.n_nodes):
					for j in range(current_layer.n_nodes):
						dw=eta * current_layer.nodes[j].delta * prev_layer.nodes[i].activation_level
						current_inputlink.arcs[j,i].update(dw)
			for i in range(len(ann.layers[-1].nodes)):
				error+= 0.5*(desiredpattern[i]-ann.layers[-1].nodes[i].activation_level)**2
	


def delta_learning(ann, trainingpairs):
	"""Delta learing algorithm, used on single layer feed-forward networks
	ann should be an ANN-object, trainingpairs a tuple of input and desired patterns. 
	Delta learning is performed on the last link in ann.links.
	"""
	
	eta=ann.learningrate
	
	
	print "####Running delta learning algorithm...####"
	for trainingpair in trainingpairs:
		rounds=0
		inputpattern, desiredpattern = trainingpair
		if VERBOSE: print "Training set: ", inputpattern, desiredpattern
		
		changed=True
		while changed and rounds<ann.trainingrounds:
			rounds+=1
			changed=False
			ann.layers[0].initialize_inputs(inputpattern)
			for l in range(len(ann.layers)):
				if VERBOSE: print "executing layer" + str(l) + "..."
				ann.layers[l].run()
				
			link=ann.links[-1]
			
			for i in range(link.prelayer.n_nodes):
				for j in range(link.postlayer.n_nodes):
					dw=eta*(desiredpattern[j] - link.postlayer.nodes[j].activation_level)*link.prelayer.nodes[i].activation_level
					if dw!=0: changed=True
					link.arcs[j,i].update(dw)
		
		
		
	
