# encoding: utf-8

"""
Methods for setting up a behavior based robot controller.
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from __future__ import with_statement
import threading
import time
import signal

from utilities import *
from image import *
from ann import *


#To avoid problems with epuck.talk responses and other annoying side-effects
#of threading:
#lock=threading.Lock()

#Prints debugging information:
VERBOSE=False

#If silent, no output is given. Should probably be left false:
SILENT=True

#height of behavior tree, 5 should be enough for now
BEHAVIOR_PRIORITY_LEVELS=5

#size of historybuffer
HISTORYLENGTH=10

#timestep for epuck movements
TIMESTEP=0.25

#decorator for using RLock on function:
def synchronized(self):
	lock = threading.RLock()
	def fwrap(function):
		def new_function(*args, **kw):
			with lock:
				return function(*args, **kw)
			return new_function
	return fwrap


def initialize_behaviors(path, epuck):
	"""Reads behavior config from text file, and returns initialized behavior controller"""
	#initialize parameter arrays:
	layers={}
	globalparams={}
	
	#get parameters from settings-file:
	try:
		f= open(path, 'r')	
		for line in f.read().split('\n'):
			if len(line)>0:
				if line.isspace() or line[0]=='#':
					continue
				else:
					line_t=line.split(':')
					#check if parameter is global or not:
					if (len(line_t)>1):
						line=line_t[1].split('=')
						#locals()[line[0]]=(eval(line[1]))
						globalparams[line[0]]=(eval(line[1]))
					else:
						params=line.split(' ')
						item={}
						for p in params:
							p=p.split('=')
							if len(p)>1:
								item[p[0]]=eval(p[1])
						if (params[0]=='layer'):
							#layername=item['layername']
							layername=params[1]
							layers[layername]=item
						else:
							print "Wrong format of ann-settings file"
							exit(1)
	except IOError as e:
		print "IO Error:", e
		sys.exit(1)
	
	behaviorlayers=[]
	lock=threading.Lock()
	threaded=globalparams['threaded']
	
	for k,l in layers.items():
		ann=None
		if l['ann']!=None:
			settings=l['ann']
			ann= create_ann(os.path.join(os.getcwd(), settings))
			#temporary fix:
			ann=load_object("oanetwork.dat")			
		behaviorlayers.append(eval(k)(epuck, ann, lock, threaded))


	#should probably be moved to somewhere else:
	epuck.set_wheel_speeds(1.0,1.0)
	B=BehaviorController(epuck, behaviorlayers, lock, threaded)
	B.initialize()
	return B


class BehaviorController(threading.Thread):
	"""Main robot controller class. All other controllers should inherit
	from this class, but should implement their own version of run().
	"""
	def __init__(self, epuck, behaviors, lock, threaded):
		if threaded: threading.Thread.__init__(self)
		#current orientation relative to starting orientation: 
		self.orientation=0.0
		self.collision_counter=0
		self.behaviors=behaviors
		self.wheel_speeds=[]
		self.next_move=[0.0]*8
		self.epuck=epuck
		self.lock=lock
		self.threaded=threaded
		self.sleeptime=0.1
		self.historybuffer=CircularBuffer(HISTORYLENGTH)
	def initialize(self):
		"""Initializes controller and behaviors. Supports parallell and 
		sequential mode.
		"""
		if self.threaded:
			for b in self.behaviors:
				b.controller=self
				b.start()
			self.start()
			#for b in self.behaviors:
				#b.join()
		else:
			#scaling factor, translates sleeptime into iteration steps. system dependent
			f=20
			for b in self.behaviors:
				b.controller=self
				#adjust sleeptime
				b.sleeptime=f*b.sleeptime
			for i in range(1000000):
				for b in self.behaviors:
					b.controller=self
					#use adjusted sleeptime to determine when to run:
					if i%(b.sleeptime)==0:
						b.run()
				self.run()
				self.sleep(0.0)
			
	def run(self):
		"""Command arbitration/fusion
		"""
		while True:
			self.lock.acquire()
			try:
				if self.next_move!=[]:
					if VERBOSE: print "Next move:", ['%3.3f' % m for m in self.next_move]
					maxval=max(self.next_move)
					for i in range(len(self.next_move)):
						if self.next_move[i]==maxval:
							dir = i
							break
					if VERBOSE: print "Moving in direction ", dir
					self.historybuffer.append(dir)
					self.epuck.move_dir(dir)
					self.next_move=[0.0]*8
	
				else:
					pass
				self.lock.release()

				if not self.threaded: break
				self.sleep(self.sleeptime)
				
			except OSError as e:
				print "Should probably not happen:", e
				sys.exit()
	def kill(self):
		"""Kill thread, and notify mother-process.
		"""
		print "Thread killed"
		try:
			self.epuck.death()
			os.kill(self.child, signal.SIGKILL)
		except OSError: pass				
	def count_collisions(self):
		"""Count the number of times one of the IR sensors is >3000
		"""
		if max(self.epuck.prox_sensor_values)>3000:
			self.collision_counter+=1
	def pause_all(self, s):
		"""behavior s pauses all others except itself
		"""
		for b in self.behaviors:
			if b!=s:
				b.active=False
	def activate_all(self):
		"""Unpause all running behaviors
		"""
		for b in self.behaviors:
			b.active=True
	def pause(self, behaviorid):
		"""Pause given behavior
		""" 
		for i in range(len(self.behaviors)):
			if self.behaviors[i].id==behaviorid:
				self.behaviors[i].active=False
				break
	def activate(self, behaviorid):
		"""Activate given behavior
		"""
		for i in range(len(self.behaviors)):
			if self.behaviors[i].id==behaviorid:
				self.behaviors[i].active=True	
				break

	def add_moves(self,new_moves):
		"""Add a list of moves to the next_moves-array
		"""
		self.next_move=map(sum,zip(new_moves, self.next_move))
	def sleep(self,sleeptime):
		if self.threaded:
			time.sleep(sleeptime)	
		else:
			self.epuck.step(self.epuck.timestep)
	def prefer_forward(self):
		"""Make small modifications to next_move in order to
		favor small changes in direction.
		"""
		mod=[0.5,0.25,0,-0.25,-0.5,-0.25,0,0.25]
		self.add_moves(mod)
			
			
			
class Behavior(threading.Thread):
	"""Behavior superclass
	"""
	def __init__(self, epuck, lock,threaded,priority):
		if threaded: threading.Thread.__init__(self)
		self.epuck=epuck
		self.controller=None
		self.priority=priority
		self.lock=lock
		self.sleeptime=1
		self.historybuffer=CircularBuffer(HISTORYLENGTH)
		
		#All behaviors start out active, with the possibility of being inhibited
		#by others
		self.active=True
		
		#If true, behavior assumes it is run as a separate thread.
		self.threaded=threaded
		
		#see comments in watch and kill-methods. 
		#Creates a child thread, which returns. The parent thread waits for a 
		#KeyboardInterrupt and then kills off the child thread:
		if self.threaded:
			self.child = os.fork()
			if self.child == 0:
				return
			else:
				self.watch()		
	def run(self):
		"""run() is implemented in child classes
		"""
		return NotImplemented
	def watch(self):
		"""watch and kill makes sure keyboard interrupts are handled by the thread. 
		Code borrowed from The Little Book of Semaphores [http://greenteapress.com/semaphores/]
		"""
		try:
			os.wait()
		except KeyboardInterrupt:
			print 'KeyboardInterrupt'
			self.kill()
		sys.exit()
	def kill(self):
		"""watch and kill makes sure keyboard interrupts are handled by the thread. 
		Code borrowed from The Little Book of Semaphores [http://greenteapress.com/semaphores/]
		"""
		print "Thread killed"
		try:
			self.epuck.death()
			os.kill(self.child, signal.SIGKILL)
		except OSError: pass
		
		
	
class Wander(Behavior):
	def __init__ (self,epuck, ann, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=BEHAVIOR_PRIORITY_LEVELS-1)
		self.ann=ann
		self.sleeptime=5
		self.id=0
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Wander"
					dir=random_int(0,7)
					if not SILENT: print "Choosen direction:", dir
					self.controller.next_move[dir]+=0.1
					if VERBOSE: print "End wander"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in Wander: ", e
				self.kill()


class AvoidObstacles(Behavior):
	def __init__ (self,epuck, ann, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=0)
		self.ann=ann
		self.status = 1
		self.sleeptime=0.1
		self.id=1
		
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Obstacle avoidance"
					prox=self.epuck.get_proximities()
					#skip rounds with empty sensor output (happens when battery level is low)
					if len(prox)>1:	
						if not SILENT: print "Prox. sensors: ", prox	
						self.ann.test([prox])
						next_move =self.ann.get_output()
						self.controller.add_moves(next_move)
						if next_move!=8*[0.0]: self.controller.prefer_forward()
						if VERBOSE: print "End obstacle avoidance"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in AvoidObstacles: ", e
				self.kill()

class GotoSound(Behavior):
	def __init__ (self,epuck, ann, lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=0.5
		self.id=2
	def empty_queue(self):
		"""Empties the queue of the receiver if length exceeds 1 item
		"""
		l=self.epuck.receiver.getQueueLength()
		if l>1:
			for i in range(l-1):
				print "emptying"
				self.epuck.receive_message()
				
	def translate(self, global_list, di=0):
		"""Translates a list of weighted moves in a global coordinate system
		into the epucks system
		"""
		local_list=8*[0.0]
		
		if di==0:
			val, bearing_angle=self.epuck.get_compass_values()
			
			di_float=bearing_angle/45.0
			di=int(di_float)
			print di, di_float
			
		for i in range(len(global_list)):
			local_list[(i+di)%8]=global_list[i]
		return local_list
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if self.epuck.getMode() == 0: # Running the simulator
						if VERBOSE: print "Go to emitter"
						self.empty_queue()
						msg, coordinates=self.epuck.receive_message()
						print self.epuck.receiver.getQueueLength()
						if coordinates != None:	
							if not SILENT: print "Message and direction: ", msg, coordinates	
							x,y,z=coordinates
							global_move_list=8*[0.0]
							if x>0.9: global_move_list[2]=1
							if x<-0.9: global_move_list[6]=1
							if z>0.9: global_move_list[4]=1
							if z<-0.9: global_move_list[0]=1
							if (x>0.0 and z>0.0): global_move_list[3]=1 
							if (x<0.0 and z<0.0): global_move_list[7]=1 
							if (x<0.0 and z>0.0): global_move_list[5]=1 
							if (x>0.0 and z<0.0): global_move_list[1]=1 
						
							moves=global_move_list
							print msg, "x:", x, "z:", z, moves
							if VERBOSE: print "Translated moves ", moves
							self.controller.add_moves(moves)
					else: # Running a real robot
						print ".."
						if VERBOSE: print "End goto sound"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GotoSound: ", e
				self.kill()
	
	
class Broadcast(Behavior):
	"""Sends regular messages to other epucks, enabling them to localize 
	sender. 
	"""
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=0.5
		self.id=6
		self.active=True
	def run(self):
		while True:
			try:
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Broadcast"
					self.epuck.send_message(time.ctime())
					if VERBOSE: print "End broadcast"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in Wander: ", e
				self.kill()

class GotoLight(Behavior):
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=2
		self.id=3
		self.light_threshold=10
		self.found_light_threshold=1000
	def run(self):
		while True:
			try:
				success=False
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Begin search for light"
					#Pause GoHome:
					self.controller.pause(4)
					
					#stop robot
					self.epuck.stop()
					#rotate 360 degrees in 8 steps:
					for i in range(8):
						im=self.epuck.snapshot()
						#returns list with length 3:
						white_pixels=pixel_count_band(im,0,30,13)
						if not SILENT: print "Goto light pixels: ", white_pixels
						
						if sum(white_pixels)>self.found_light_threshold: 
							if not SILENT: print "Found light!"
							if not SILENT: im.show()
							print "Found light!"
							success=True
							break
						elif max(white_pixels) > self.light_threshold:
							if not SILENT: print "Moving towards light..."
							self.controller.next_move[0]+=1
							break
						else:
							self.epuck.spin_angle(45)
					#pause self, activate GoHome:
					if success:
						self.epuck.wait(2)
						self.controller.pause(3)
						self.controller.activate(4)
					self.epuck.set_wheel_speeds(0.5,0.5)
					
					if VERBOSE: print "End search for light"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GotoLight: ", e
				self.kill()

class GoHome(Behavior):
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		#this behavior should start out inactive
		self.active=False
		self.status = 1
		self.sleeptime=2
		self.id=4
		self.redness_threshold=30
		self.found_home_threshold=1000
	def run(self):
		while True:
			try:
				success=False
				if self.active:
					self.lock.acquire()
					if VERBOSE: print "Begin Go home"
					#Pause GotoLight:
					self.controller.pause(3)
					
					#stop robot
					self.epuck.stop()
					
					#rotate epuck 360 degrees in 8 steps:
					for i in range(8):
						im=self.epuck.snapshot()
						#returns list with length 3:
						red_pixels=redness_count_band(im,0,30,13)
						if not SILENT: print "Go home pixels: ", red_pixels
						if sum(red_pixels)>self.found_home_threshold: 
							if not SILENT: print "Found home!"
							#self.epuck.send_message(time.ctime())
							if not SILENT: im.show()
							print "Found home!"
							success=True
							break
						elif max(red_pixels)> self.redness_threshold:
							if not SILENT: print "Moving towards redness..."
							#self.epuck.send_message(time.ctime())
							print  "Moving towards redness..."
							self.controller.next_move[0]+=1
							if red_pixels[0]>10:
								self.controller.next_move[7]+=1
							if red_pixels[2]>10:
								self.controller.next_move[1]+=1
							if red_pixels[1]>10:
								self.controller.next_move[0]+=1
							break
						else:
							self.epuck.spin_angle(45)
							
					#pause self, activate GotoLight:
					if success:
						self.epuck.wait(2)
						self.controller.pause(4)
						#self.controller.pause(2)
						#self.controller.pause(1)
						#self.controller.pause(0)
						#self.controller.activate(6)
						self.controller.activate(3)
						#self.epuck.set_wheel_speeds(0.0,0.0)
					self.epuck.set_wheel_speeds(0.5,0.5)
					
					if VERBOSE: print "End Go home"
					self.lock.release()
				
				if not self.threaded: break
				self.controller.sleep(self.sleeptime)
			except ValueError as e:
				print "ValueError in GoHome: ", e
				self.kill()
	
	
	
#todo..			
class FollowLine(Behavior):
	"""incomplete..
	"""
	def __init__ (self,epuck, ann,lock,threaded):
		Behavior.__init__(self, epuck, lock,threaded,priority=3)
		self.ann=ann
		self.status = 1
		self.sleeptime=2
		self.id=5
	def run(self):
		return NotImplemented
