# encoding: utf-8

from controller import *
import math
import time
import sys,os
import Image

from ann import *
from utilities import *
from behavior_layers import *

VERBOSE=False
SILENT=True

class Epuck_Webots (DifferentialWheels):
	def __init__(self, tempo = 1.0):
		DifferentialWheels.__init__(self)
		
		##PHYSICAL EPUCK PROPERTIES:##
		self.max_wheel_speed = 1000
		#Number of proximity sensors:
		self.n_prox_sensors = 8
		#for wheel encoders:
		#self.encoder_resolution = 159.23
		self.encoder_resolution = 159.23
		#Camera resolution:
		self.camera_xres = 40
		self.camera_yres = 40
		#from EPuckSound.proto:
		self.wheel_radius = 0.0205 #m
		self.wheel_diameter=self.wheel_radius*2
		self.axle_length = 0.052 #m
		
		#Real-time seconds per timestep:
		self.timestep_duration = 1
		
		#Upper velocity bound = Fraction of the robot's maximum velocity = 1000 = 1 wheel revolution/sec
		self.tempo = tempo
		self.wheel_speed_scaling=self.max_wheel_speed*self.tempo
		#self.max_spin_rate = self.tempo * (self.wheel_diameter / self.axle_length) 
		self.left_speed=0.0
		self.right_speed=0.0
	
		#Fetch from worldinfo.basicTimeStep:
		self.timestep = int(self.getBasicTimeStep())
		
		#Setup sensors:
		self.enableEncoders(self.timestep)
		self.camera = self.getCamera('camera')
		self.camera.enable(4*self.timestep)
		self.prox_sensor_values = [0.0]*self.n_prox_sensors
		self.prox_sensors = [self.getDistanceSensor('ps'+str(x)) for x in range(self.n_prox_sensors)]
		map((lambda s: s.enable(self.timestep)), self.prox_sensors) # Enable all distance sensors
		
		#So that the keyboard can be used to control epuck:
		self.keyboardEnable(6*self.timestep)
		
		#Add emitter and receiver to epuck.
		self.emitter=self.getEmitter("emitter")
		self.receiver=self.getReceiver("receiver")
		self.receiver.enable(4*self.timestep)
		
		#compass
		self.compass=self.getCompass('compass')
		self.compass.enable(self.timestep)
		
	def wait(self, seconds = 1.0):
		"""Set epuck to sleep for a given number of seconds. Will pause entire
		simulation, so should be used with caution. 
		"""
		print "waiting..."
		time.sleep(seconds)
	
	def stop(self):
		"""Stop wheel movement
		"""
		self.set_wheel_speeds(0,0)

	def set_wheel_speeds(self,left = 0.0, right = 0.0):
		"""Sets wheel speeds. left and right are floats in the range [-1,1],
		and is scaled by self.wheel_speed_scaling
		"""
		s = self.wheel_speed_scaling
		left=max(-1.0, min(1.0, left))
		right=max(-1.0, min(1.0, right))
		
		self.left_speed=left
		self.right_speed=right
		self.setSpeed(int(left*s),int(right*s))
	
	def change_wheel_speeds(self,dl=0.0,dr=0.0,t=0):
		"""Changes wheel speeds by factors dl and dr. If t is 0, change is permanent,
		else it last for a duration of t seconds. 
		"""
		old_left, old_right=self.left_speed, self.right_speed
		new_left=self.left_speed+dl
		new_right=self.right_speed+dr
		self.set_wheel_speeds(new_left,new_right)
		
		
		if t!=0:
			time=0.0
			
			while time < t:
				self.step(self.timestep)
				time+=self.timestep/1000.0
			self.set_wheel_speeds(old_left,old_right)
	
	def move_dir(self, dir=0):
		"""Move in direction (relative to current orientation):
		7 0 1   FRONT
		5 X 2     ^
		5 4 3     | 
	    
		angle a, dir d:
		d*45=a
		""" 	
		angle=dir*45
		self.spin_angle(angle)
		if VERBOSE: print "Spinning ", angle, " degrees." 
		
	def move_absolute_dir(self, dir=0):
		"""Similar to move_dir, but dir is in global (world) coordinates.
		Compass is used to find the current orientation of the e-puck. 

			7 0 1   	x
			5 X 2     ^
			5 4 3     | 
		"""
		val, bearing_angle=self.get_compass_values()
		spin_angle=bearing_angle+dir*45
		self.spin_angle(spin_angle)
		if VERBOSE: print "Spinning ", spin_angle, " degrees." 
	
	def spin_angle(self, angle=0.0):
		"""Spin a given angle (in degrees). The angle is converted to wheel
		encoder units. There are encoderResolution units per radians of wheel rotation
		"""
		old_left, old_right=self.left_speed, self.right_speed
		rot_angle = abs(angle)*math.pi/180.0
		
		if angle > 180:
			angle = angle - 360
		
		if angle < 0:
			self.set_wheel_speeds(-1,1)
		else:
			self.set_wheel_speeds(1,-1)
			
		self.setEncoders(0.0,0.0)
		
		right_encoder=self.getRightEncoder()
		left_encoder=self.getLeftEncoder()
		
		total_encoder_steps = rot_angle*self.axle_length/self.wheel_radius
		
		current_step = abs(self.getLeftEncoder())
		current_right_step = self.getRightEncoder()
		stop_step = current_step + total_encoder_steps
		#self.setEncoders(stop_step,current_step)
		#print "enc", stop_step, current_step
		
		self.setEncoders(0.0,0.0)
		
		#slightly ugly, temporary fix:
		const=-50
		total_encoders = rot_angle * self.encoder_resolution * self.axle_length / self.wheel_diameter
		current_encoder = abs(self.getLeftEncoder())
		stop_encoder = current_encoder + total_encoders + const
		
		
		#print "total,stop,curr", total_encoders, stop_encoder, current_encoder
		while current_encoder < stop_encoder:
			current_encoder = abs(self.getLeftEncoder())
			self.step(self.timestep)
			

		
		self.set_wheel_speeds(old_left,old_right)

	def get_compass_values(self):
		"""Return compass coordinates and bearing angle (relative to 
		positive x-dir)
		"""
		val=self.compass.getValues()
		angle_rad = math.atan2(val[2], val[0])
		bearing = (angle_rad - 1.5708) / math.pi * 180.0+180.0
		return val, bearing


	def get_proximities(self):
		"""Store and return proximity sensor values, converted to integers.
		"""
		for i in range(self.n_prox_sensors):
			self.prox_sensor_values[i] = int(self.prox_sensors[i].getValue())
		return self.prox_sensor_values

	def snapshot(self):
		"""Aquire string-image from epuck camera, and convert it to python image (using PIL)
		"""
		strImage=self.camera.getImage()
		im = Image.fromstring('RGB',(self.camera.getWidth(), self.camera.getHeight()), strImage)
		return im
	
	def send_message(self, msg):
		"""Send message using emitter.
		"""
		self.emitter.send(msg)
	
	def receive_message(self):
		"""Receive and return message using receiver
		"""
		msg=None
		dir=None
		if self.receiver.getQueueLength()!=0:
			dir=self.receiver.getEmitterDirection()
			msg=self.receiver.getData()
			self.receiver.nextPacket()
			if VERBOSE: print "Message received:", msg
		else:
			if VERBOSE: print "No message received"
		return msg, dir

	def interactive(self):
		"""Control epuck interactively, using the keyboard. 
		"""
		try:
			#img-counter
			i=0
			while True:
				c = self.keyboardGetKey()
				if c==ord('S'):
					im=self.snapshot()
					print "saving image..."
					imgname="snapshot" + str(i) + ".jpg"
					im.save(imgname)
					i+=1
				elif c==ord('Q'):
					break
				elif c==ord('P'):
					print self.get_proximities()
				elif c==ord('E'):
					print "emitting.."
					self.emitter.send(time.ctime())
				elif c==ord('R'):
					print "todo: getting receiver data"
					if self.receiver.getQueueLength()!=0:
						msg=self.receiver.getData()
						x,y,z=self.receiver.getEmitterDirection()
						print msg, "x:", x, "z:", z
					else:
						print "nothing to receive"
					#print self.get_mic_amplitudes()
				elif c==ord('C'):
					print "compass values", self.get_compass_values()
				elif c==Robot.KEYBOARD_RIGHT:
					self.spin_angle(45)
				elif c==Robot.KEYBOARD_LEFT:
					self.spin_angle(-45)
				elif c==ord('X'):
					#reorient to positive x-direction
					self.move_absolute_dir(2)
				elif c==Robot.KEYBOARD_UP:
					self.change_wheel_speeds(0.25,0.25)
				elif c==Robot.KEYBOARD_DOWN:
					self.change_wheel_speeds(-0.25,-0.25)
				
				if self.step(self.timestep) == -1: break
		except KeyboardInterrupt as e:
			print "Quitting..: ", e
		finally:
			pass

		
def robot_train_oa(ann):
	"""Delta learning training data for obstacle
	avoidance on e-puck.
	"""
	front1=([0,0,0,0,0,0,0,300],[0,-1])
	front2=([300,0,0,0,0,0,0,0],[-1,0])
	right1=([0,300,0,0,0,0,0,0],[-1,0])
	right2=([0,0,300,0,0,0,0,0],[-1,0])
	#right3=([0,0,0,300,0,0,0,0],[-1,0])
	left1=([0,0,0,0,0,0,300,0],[0,-1])
	left2=([0,0,0,0,0,300,0,0],[0,-1])
	#left3=([0,0,0,0,300,0,0,0],[0,-1])
	
	delta_learning(ann, [front1,front2,right1,right2,left1, left2])	


