#!/usr/bin/env python
# encoding: utf-8

"""
Image processing methods.
2009, Maria Johansen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


#Image class from PIL:
import Image as PImage
import math

#Print debugging information:
VERBOSE=False

RGB_COLORS = ["red","green","blue"]

def normalize(image):
	pixels=image.load()
	width,height=image.size
	for x in range(0,width):
		for y in range(0,height):
			r,g,b=pixels[x,y]
			pixels[x,y]=normalize_pixel(r,g,b)
	
	
def pixel_count_band(image,bandstart,bandend,boxwidth,threshold=400):
	"""Return list of # pixels above threshold for each part of band
	"""
	try:
		width,height=image.size
		startx=0
		
		starty=bandstart
		endy=bandend
			
		pixels=image.load()
		pixelcounts = []
		
		pixelcount=0
		
		for a in range(0,width/boxwidth):
			for x in range(a*boxwidth,a*boxwidth+boxwidth):
				for y in range(starty,endy):
					r,g,b=pixels[x, y]
					if pixel_intensity(r,g,b)>threshold:
						pixelcount+=1
			pixelcounts.append(pixelcount)
			pixelcount=0
		return pixelcounts
	except OSError as e:
		print "OSError:", e	
			
def redness_count_band(image, bandstart,bandend,boxwidth, rg_threshold=1.3):
	"""Count the number of pixels with a 'redness' above threshold for each part 
	of band in image.
	"""
	try:
		width,height=image.size
		startx=0
	
		starty=bandstart
		endy=bandend
			
		pixels=image.load()
		pixelcounts = []
		
		pixelcount=0
		
		#a red:green-ratio of >=1.3 is a fairly good indicator of red and brown colors. 
		#rg_threshold=1.3
		
		for a in range(0,width/boxwidth):
			for x in range(a*boxwidth,a*boxwidth+boxwidth):
				for y in range(starty,endy):
					r,g,b=pixels[x, y]
					#add 1 to r and g to avoid zero-division-errors
					if (r+1.0)/(g+1.0) >=rg_threshold:
						pixelcount+=1
			pixelcounts.append(pixelcount)
			pixelcount=0
		return pixelcounts
	except OSError as e:
		print "OSError:", e	

def threshold_mask_intensity(image,threshold=400):
	"""Return part of image with intensity greater that threshold. 
	For visual purposes, the thresholded image is colored in a contrast color,
	while the background is equal to the original image.
	(note: will overwrite original image object)
	"""
	width,height=image.size
	pixels=image.load()
	
	pixelcount=0
	
	for x in range(0,width):
		for y in range(0,height):
			r,g,b=pixels[x, y]
			if pixel_intensity(r,g,b)>threshold:
				pixelcount+=1
				pixels[x, y]=(255,0,0)
	return pixelcount
			

def threshold_mask_rgratio(image, threshold=1.3):
	"""Return part of image with red:green ratio equal to or greater than
	threshold. For visual purposes, the thresholded part of the image is colored
	in a contrast color, while the background is equal to the original image.
	(note: will overwrite original image object)
	"""
	width,height=image.size
	pixels=image.load()
	
	pixelcount=0
	
	for x in range(0,width):
		for y in range(0,height):
			r,g,b=pixels[x, y]
			if ((r+1.0)/(g+1.0) >= threshold):
				pixelcount+=1
				pixels[x, y]=(0,255,0)
	return pixelcount
	
def threshold_mask_colormax(image, color='r'):
	"""Return part of image with pixels where 'color' is the highest color 
	component. 
	For visual purposes, the thresholded part of the image is colored
	in a contrast color, while the background is equal to the original image.
	(note: will overwrite original image object)
	"""
	width,height=image.size
	pixels=image.load()
	
	pixelcount=0
	
	for x in range(0,width):
		for y in range(0,height):
			r,g,b=pixels[x, y]
			if color=='r':
				if (r>g and r>b):
					pixelcount+=1
					pixels[x, y]=(0,255,0)
			elif color=='g':
				if (g>r and g>b):
					pixelcount+=1
					pixels[x, y]=(0,255,0)			
			elif color=='b':
				if (b>g and b>r):
					pixelcount+=1
					pixels[x, y]=(0,255,0)				
	return pixelcount

def intensity_band(image,bandstart,bandend, boxwidth):
	"""Return list of avg intensities in a center band of image. height and 
	width specifies the size of averaging mask. Currently, no error checking
	is done. 
	"""
	try:
		width,height=image.size
		y_mid=height/2
		startx=0
		
		starty=bandstart
		endy=bandend
		
		
		pixels=image.load()
		avg_intensities = []
		boxheight=bandend-bandstart
		
		avg_intensity=0
		for a in range(0,width/boxwidth):
			for x in range(a*boxwidth,a*boxwidth+boxwidth):
				for y in range(starty,endy):
					r,g,b=pixels[x, y]
					avg_intensity+=pixel_intensity(r,g,b)
			avg_intensities.append(avg_intensity/(boxheight*boxwidth))
			avg_intensity=0
		return avg_intensities
	except OSError as e:
		print "OSError:", e
	
	
def normalize_pixel(r,g,b,threshold=50):
	"""Return normalized pixel as tuple (r,g,b)
	"""

	if pixel_intensity(r,g,b)<threshold:
		f=0	
	else:
		m=max(r,g,b)
		print m
		f=255/m
	print (r*f, g*f, b*f)
	return (r*f, g*f, b*f)
	
def pixel_intensity(r,g,b):
	"""Calculate the intensity of pixel with
	values r,g,b
	"""
	return math.sqrt(r*r + b*b + g*g)
	
